from opennyai import ner, utils
from opennyai.pipeline import Pipeline
from opennyai.rhetorical_roles.rhetorical_roles import RhetoricalRolePredictor
from opennyai.summarizer.ExtractiveSummarizer import ExtractiveSummarizer

__version__ = "{0}.{1}.{2}".format("0", "0", "14")
__author__ = 'OpenNyAI Team'
__credits__ = 'Opennyai'
